{*
Camera unit

@author         Philipp Hartl
@version        0.1
@license
This source file is part of BotanicalVis,
a tool to visualize huge hierarchies in 3D-space

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
*}

unit Camera;

interface

uses
  dglOpenGL, Geo;

type

  /// camera projection types
  TProjectionType = (PT_ORTHOGRAPHIC, PT_PERSPECTIVE);

  {*
  camera
  for manipulating viewing and objects

  cameras coordinate system (right handed)
  x ... strafe left(-)/right(+)
  y ... up(+)/down(-)
  z ... forward(-)/backward(+)
  global means the origin of the scene (0,0,0) and its coordinate system
  *}
  TCamera = class(TObject)
  private
    FFovy, FAspect, FNear, FFar: TGLfloat;
    FPos, FView, FUp, FRight: TVector3s;
    FPT: TProjectionType;
    FTransform: TMatrix16f;     /// OpenGL col-major matrix

  public
    constructor Create(pos, dest, up: TVector3f);
    destructor Destroy; override;

    procedure LookAt;           /// calculate current camera matrix
    procedure RotateGlobal(const alpha, x, y, z: TGLfloat);
    procedure RotateLocal(const alpha, x, y, z: TGLfloat);
    procedure TranslateGlobal(const dist, x, y, z: TGLfloat);
    procedure TranslateLocal(const dist, x, y, z: TGLfloat);
//    procedure ZoomAll;          /// a zoom extends function

    property Fovy: TGLfloat read FFovy write FFovy;
    property AspectRatio: TGLfloat read FAspect write FAspect;
    property NearPlane: TGLfloat read FNear write FNear;
    property FarPlane: TGLfloat read FFar write FFar;
    property ProjectionType: TProjectionType read FPT write FPT;
  end;

implementation

//------------------------------------------------------------------------------
// TCamera
//------------------------------------------------------------------------------

{*
second constructor

@param  pos, the starting camera position
@param  dest, the destination point the camera should look like
@param  up, the up vector of the camera (should be perpendicular to viewing-vector)
*}
constructor TCamera.create(pos, dest, up: TVector3s);
begin
  FPos:= pos;
  FView:= Vector3fNormalize(Vector3fSub(dest, pos));
  FUp:= up;
  FRight:= Vector3fNormalize(Vector3fCrossProduct(FView, FUp));

  // OpenGL col-major format
  FTransform:= Matrix16f_Identity;
  Matrix16fSetVector(FTransform, FRight, 0);
  Matrix16fSetVector(FTransform, FUp, 4);
  Matrix16fSetVector(FTransform, Vector3fNegate(FView), 8);     // right handed
  Matrix16fSetVector(FTransform, FPos, 12);

  FFovy:= 60.0;
  FAspect:= 4/3;
  FNear:= 0.1;
  FFar:= 1000.0;
  FPT:= PT_PERSPECTIVE;
end;

{*
destructor
*}
destructor TCamera.Destroy;
begin
  inherited Destroy;
end;

{*
rotate the objects around their coordinate axes (global)

@param  alpha, the angle amount for rotation
@param  x, y, z, the axes and direction of rotation
*}
procedure TCamera.RotateGlobal(const alpha, x, y, z: TGLfloat);
var m: TMatrix16f;
begin
  // R ... inverse = transpose for R, watch out for T
  m:= Matrix16fTranspose(FTransform);
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix;
  glLoadMatrixf(PGLfloat(@m));
  glRotatef(alpha, x, y, z);
  glGetFloatv(GL_MODELVIEW_MATRIX, PGLfloat(@m));
  glPopMatrix;
  FTransform:= Matrix16fTranspose(m);
end;

{*
rotate around local coordinate system (local)

@param  alpha, the angle amount for rotation
@param  x, y, z, the axes and direction of rotation
*}
procedure TCamera.RotateLocal(const alpha, x, y, z: TGLfloat);
begin
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix;
  glLoadMatrixf(PGLfloat(@FTransform));
  glRotatef(alpha, x, y, z);
  glGetFloatv(GL_MODELVIEW_MATRIX, PGLfloat(@FTransform));
  glPopMatrix;
end;

{*
translate the objects in global coordinate system

@param  dist, the moving distance
@param  x, y, z, the axes and direction to move
*}
procedure TCamera.TranslateGlobal(const dist, x, y, z: TGLfloat);
begin
  FTransform[12]:= x * dist + FTransform[12];
  FTransform[13]:= y * dist + FTransform[13];
  FTransform[14]:= z * dist + FTransform[14];
end;

{*
translate the camera in its local coordinate system

@param  distance to move
@param  x, y, z, the axes and direction
*}
procedure TCamera.TranslateLocal(const dist, x, y, z: TGLfloat);
var dx, dy, dz: TGLfloat;
begin
  // move local to global
  dx:= x * FTransform[0] + y * FTransform[4] + z * FTransform[8];
  dy:= x * FTransform[1] + y * FTransform[5] + z * FTransform[9];
  dz:= x * FTransform[2] + y * FTransform[6] + z * FTransform[10];
  translateGlobal(dist, dx, dy, dz);
end;

{*
update camera matrix for the scene, similar to gluLookAt
call this at first in MODELVIEW stage after initialising with identity matrix
*}
procedure TCamera.LookAt;
var m: TMatrix16f;
begin
  m:= Matrix16f_Identity;

  // R^-1
  m[0]:= FTransform[0]; m[1]:= FTransform[4]; m[2]:= FTransform[8];
  m[4]:= FTransform[1]; m[5]:= FTransform[5]; m[6]:= FTransform[9];
  m[8]:= FTransform[2]; m[9]:= FTransform[6]; m[10]:= FTransform[10];

  // R*T
  m[12]:= -(FTransform[0] * FTransform[12] + FTransform[1] * FTransform[13] + FTransform[2]  * FTransform[14]);
  m[13]:= -(FTransform[4] * FTransform[12] + FTransform[5] * FTransform[13] + FTransform[6]  * FTransform[14]);
  m[14]:= -(FTransform[8] * FTransform[12] + FTransform[9] * FTransform[13] + FTransform[10] * FTransform[14]);

  glLoadMatrixf(PGLfloat(@m));
end;


end.