{*
geometry (vector, matrix) stuff
matrices are in row-major format!

@author         Philipp Hartl
@version        0.1
@license
This source file is part of BotanicalVis,
a tool to visualize huge hierarchies in 3D-space

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
*}

unit Geo;

interface

uses
   SysUtils,
   dglOpenGL;

type

  TColor3f = TVector3f;
  TMatrix16f = array[0..15] of TGLfloat;

  /// vector
  function Vector3f(const x,y,z: TGLfloat): TVector3f;
  function Vector3fAdd(const v1, v2: TVector3f): TVector3f;
  function Vector3fAsString(const v: TVector3f): String;
  function Vector3fCrossProduct(const v1, v2: TVector3f): TVector3f;
  function Vector3fDotProduct(const v1, v2: TVector3f): TGLfloat;
  function Vector3fLength(const v: TVector3f): TGLfloat;                /// Sqrt(v)
  /// return a non-homogenous coloumn vector of matrix
  function Vector3fOfMatrix4f(const m: TMatrix4f; const col: TGLushort): TVector3f;
  function Vector3fNegate(const v: TVector3f): TVector3f;
  function Vector3fNormalize(const v: TVector3f): TVector3f;
  function Vector3fScale(const v: TVector3f; const x: TGLfloat): TVector3f;
  function Vector3fSLength(const v: TVector3f): TGLfloat;               /// x+y+z
  function Vector3fSub(const v1, v2: TVector3f): TVector3f;

  /// matrix
  function Matrix4fTranspose(const m: TMatrix4f): TMatrix4f;
  procedure Matrix4fSetCol(var m: TMatrix4f; const v: TVector3f; const col: TGLushort);
  procedure Matrix4fSetRow(var m: TMatrix4f; const v: TVector3f; const row: TGLushort);
  function Matrix16fTranspose(const m: TMatrix16f): TMatrix16f;
  procedure Matrix16fSetVector(var m: TMatrix16f; const v: TVector3f; const startindex: TGLushort);
  function Matrix16fPrint(const m: TMatrix16f): String;

  // color
  function Color3f(const r, g, b: TGLfloat): TColor3f;
  function Color3bToColor3f(const r, g, b: TGLubyte): TColor3f;


const
  Vector3f_Zero: TVector3f   = (0, 0, 0);
  Vector3f_UnitX: TVector3f  = (1, 0, 0);
  Vector3f_UnitY: TVector3f  = (0, 1, 0);
  Vector3f_UnitZ: TVector3f  = (0, 0, 1);
  Vector3f_UScale: TVector3f = (1, 1, 1);

  Matrix4f_Identity: TMatrix4f   = ((1,0,0,0),
                                    (0,1,0,0),
                                    (0,0,1,0),
                                    (0,0,0,1));

  Matrix4f_Zero: TMatrix4f       = ((0,0,0,0),
                                    (0,0,0,0),
                                    (0,0,0,0),
                                    (0,0,0,0));

  Matrix16f_Identity: TMatrix16f = (1,0,0,0,
                                    0,1,0,0,
                                    0,0,1,0,
                                    0,0,0,1);

  Color3f_Black: TColor3f = (0, 0, 0);
  Color3f_White: TColor3f = (1, 1, 1);

var
  ColorCylinder, ColorSphere, ColorUnknown: TColor3f;

implementation

//------------------------------------------------------------------------------
// color
//------------------------------------------------------------------------------

function Color3f(const r, g, b: TGLfloat): TColor3f;
begin
  result[0]:= r;
  result[1]:= g;
  result[2]:= b;
end;

function Color3bToColor3f(const r, g, b: TGLubyte): TColor3f;
begin
  result[0]:= r / 255;
  result[1]:= g / 255;
  result[2]:= b / 255;
end;

//------------------------------------------------------------------------------
// Vector
//------------------------------------------------------------------------------

function Vector3f(const x,y,z: TGLfloat): TVector3f;
begin
  result[0]:= x;
  result[1]:= y;
  result[2]:= z;
end;

function Vector3fAsString(const v: TVector3f): String;
begin
  result:= 'x: ' + FloatToStr(v[0]) + ' y: ' + FloatToStr(v[1]) + ' z: ' + FloatToStr(v[2]);
end;

function Vector3fNegate(const v: TVector3f): TVector3f;
begin
  result[0]:= -v[0];
  result[1]:= -v[1];
  result[2]:= -v[2];
end;

function Vector3fAdd(const v1, v2: TVector3f): TVector3f;
begin
  result[0]:= v1[0] + v2[0];
  result[1]:= v1[1] + v2[1];
  result[2]:= v1[2] + v2[2];
end;

function Vector3fScale(const v: TVector3f; const x: TGLfloat): TVector3f;
begin
  result[0]:= v[0] * x;
  result[1]:= v[1] * x;
  result[2]:= v[2] * x;
end;

function Vector3fLength(const v: TVector3f): TGLfloat;
begin
  result:= sqrt(Vector3fSLength(v));
end;

function Vector3fSLength(const v: TVector3f): TGLfloat;
begin
  result:= sqr(v[0]) + sqr(v[1]) + sqr(v[2]);
end;

function Vector3fNormalize(const v: TVector3f): TVector3f;
var x: TGLfloat;
begin
  x:= Vector3fLength(v);
  if x = 0 then result:= Vector3f_Zero
  else result:= Vector3fScale(v, (1/x));
end;

function Vector3fCrossProduct(const v1, v2: TVector3f): TVector3f;
begin
  result[0]:= v1[1] * v2[2] - v1[2] * v2[1];
  result[1]:= v1[2] * v2[0] - v1[0] * v2[2];
  result[2]:= v1[0] * v2[1] - v1[1] * v2[0];
end;

function Vector3fDotProduct(const v1, v2: TVector3f): TGLfloat;
begin
  result:= v1[0] * v2[0] + v1[1] * v2[1] + v1[2] * v2[2];
end;

function Vector3fOfMatrix4f(const m: TMatrix4f; const col: TGLushort): TVector3f;
begin
  result:= Vector3f(m[0,col], m[1,col], m[2,col]);
end;

function Vector3fSub(const v1, v2: TVector3f): TVector3f;
begin
  result[0]:= v1[0] - v2[0];
  result[1]:= v1[1] - v2[1];
  result[2]:= v1[2] - v2[2];
end;

//------------------------------------------------------------------------------
// Matrix
//------------------------------------------------------------------------------

function Matrix4fTranspose(const m: TMatrix4f): TMatrix4f;
var i,j: shortint;
begin
  for i:=0 to 3 do
    for j:= 0 to 3 do
      result[j,i]:= m[i,j];
end;

function Matrix16fTranspose(const m: TMatrix16f): TMatrix16f;
begin
  result[0]:= m[0]; result[1]:= m[4]; result[2]:= m[8]; result[3]:= m[12];
  result[4]:= m[1]; result[5]:= m[5]; result[6]:= m[9]; result[7]:= m[13];
  result[8]:= m[2]; result[9]:= m[6]; result[10]:= m[10]; result[11]:= m[14];
  result[12]:= m[3]; result[13]:= m[7]; result[14]:= m[11]; result[15]:= m[15];
end;

procedure Matrix4fSetCol(var m: TMatrix4f; const v: TVector3f; const col: TGLushort);
begin
  m[0,col]:= v[0];
  m[1,col]:= v[1];
  m[2,col]:= v[2];
end;

procedure Matrix4fSetRow(var m: TMatrix4f; const v: TVector3f; const row: TGLushort);
begin
  m[row,0]:= v[0];
  m[row,1]:= v[1];
  m[row,2]:= v[2];
end;

procedure Matrix16fSetVector(var m: TMatrix16f; const v: TVector3f; const startindex: TGLushort);
begin
  m[startindex]:= v[0];
  m[startindex + 1]:= v[1];
  m[startindex + 2]:= v[2];
end;

function Matrix16fPrint(const m: TMatrix16f): String;
var i: TGLushort;
begin
  for i:= 0 to 15 do
    result:= result + ' ' + FloatToStr(m[i]);
end;

end.