{*
Main unit
all process starts here and should end here

@author         Philipp Hartl
@version        0.1
@license
This source file is part of BotanicalVis,
a tool to visualize huge hierarchies in 3D-space

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
*}
unit Main;

interface

uses
  Windows, Classes, Controls, Forms, StdCtrls, ShlObj, ActiveX, ExtCtrls, SysUtils, IniFiles,
  Logger, Rendering, Data, ComCtrls, Graphics, Dialogs, Geo, dglOpenGL;

type

  PTTreeView = ^TTreeView;
  TReadDirThread = class;

  {*
  main formular, all settings and handling are here
  *}
  TMainFrm = class(TForm)
    GroupBox1: TGroupBox;
    edtDir: TEdit;
    btnScan: TButton;
    GroupBox2: TGroupBox;
    lvColTyp: TListView;
    colDlg: TColorDialog;
    Label1: TLabel;
    Label2: TLabel;
    edAlpha: TEdit;
    edBeta: TEdit;
    Label3: TLabel;
    lblInfo: TLabel;
    GroupBox3: TGroupBox;
    tvData: TTreeView;

    procedure btnScanClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    function GetColor3fToTyp(const t: String): TColor3f;
    procedure lvColTypKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure lvColTypCustomDrawSubItem(Sender: TCustomListView;
      Item: TListItem; SubItem: Integer; State: TCustomDrawState;
      var DefaultDraw: Boolean);
    procedure lvColTypColumnClick(Sender: TObject; Column: TListColumn);
    procedure lvColTypDblClick(Sender: TObject);
    procedure tvDataClick(Sender: TObject);

  private
    FReadDirThread: TReadDirThread;
    FReadDirThreadRunning: Boolean;             /// is FReadDirThread running?
    FCntFiles, FCntDirs: Integer;
    FProgDir: String;

    function GetDirectory(caption: string = 'please select directory'): String;
    procedure ReadDirThreadFinished(Sender: TObject);
    procedure ReadIniSettings;
    function SetRotAngle(angle: PGLfloat; edt: TEdit): boolean;
    procedure SetGlobalColors();                /// ColorSphere, Stem, Unknown
    procedure WriteIniSettings;
  end;

  {*
   read directory data thread
   this thread walks through a directory structure recursively
   and build up a corresponding tree structure
   *}
  TReadDirThread = class(TThread)
  private
    FDir: String;
    FPDI: PTDirectoryItem;              /// pdi structure
    FPTV: PTTreeView;

    procedure BuildRenderData;
    procedure UpdateCurrentDir;
    procedure FreeDI(pdi: PTDirectoryItem);
    function GetFiles(const dir: String; const parent: TTreeNode): PTDirectoryItem;

  protected
    procedure Execute; override;

  public
    constructor Create(CreateSuspended: Boolean; dir: String; const ptv: PTTreeView);
    destructor Destroy(); override;
    property Dir: String write FDir;
  end;

  /// some color conversion
  function HexToTColor(color: String): TColor;
  function TColorToHex(color: TColor): String;
  function TColorToTRGBQuad(color: TColor): TRGBQuad;
  function TRGBQuadToColor3f(color: TRGBQuad): TColor3f;

var
  MainFrm: TMainFrm;

implementation

{$R *.dfm}

//------------------------------------------------------------------------------

function TColorToHex(color: TColor): String;
begin
  result:= IntToHex(GetRValue(color), 2) +
           IntToHex(GetGValue(color), 2) +
           IntToHex(GetBValue(color), 2);
end;

function HexToTColor(color: String): TColor;
begin
  result:= RGB(StrToInt('$' + Copy(color, 1, 2)),
               StrToInt('$' + Copy(color, 3, 2)),
               StrToInt('$' + Copy(color, 5, 2)));
end;

function TColorToTRGBQuad(color: TColor): TRGBQuad;
begin
  result.rgbBlue:=  color and $000000FF;
  result.rgbGreen:= color and $0000FF00 shr 8;
  result.rgbRed:=   color and $00FF0000 shr 16;
end;

function TRGBQuadToColor3f(color: TRGBQuad): TColor3f;
begin
  result:= Color3bToColor3f(color.rgbBlue, color.rgbGreen, color.rgbRed);
end;


//------------------------------------------------------------------------------
// TMainFrm
//------------------------------------------------------------------------------

{*
check rotation angle
*}
function TMainFrm.SetRotAngle(angle: PGLfloat; edt: TEdit): boolean;
begin
  result:= false;
  try
    angle^:= StrToFloat(edt.Text);
    if (angle^ >= 0) and (angle^ <= 360) then result:= true
    else
    begin
      MessageBox(0,PChar('Angle out of range!' + #13#10 + '0 <= angle <= 360'), 'Rotation', MB_OK or MB_ICONINFORMATION);
      edt.SetFocus;
    end;
  except
    on e: EConvertError do
    begin
      MessageBox(0,PChar('Not valid float value!' + #13#10 + '[0-9|,]'), 'Rotation', MB_OK or MB_ICONERROR);
      edt.SetFocus;
    end;
  end;
end;

{*
set global color values
*}
procedure TMainFrm.SetGlobalColors;
var
  i: integer;
begin
  for i:= 0 to lvColTyp.Items.Count-1 do
  begin
    with lvColTyp.Items.Item[i] do
    begin
      if Caption = 'sphere' then ColorSphere:= TRGBQuadToColor3f(TColorToTRGBQuad(HexToTColor(SubItems.Strings[0])))
      else if Caption = 'stem' then ColorCylinder:= TRGBQuadToColor3f(TColorToTRGBQuad(HexToTColor(SubItems.Strings[0])))
      else if Caption = 'unknown' then ColorUnknown:= TRGBQuadToColor3f(TColorToTRGBQuad(HexToTColor(SubItems.Strings[0])));
    end;
  end;
end;

{*
select directory by opening a system file dialog
on a valid dir, start recursive parsing the directory structure by a thread

@see  TReadDir thread
*}
procedure TMainFrm.btnScanClick(Sender: TObject);
var
  dir: String;
begin
  if not FReadDirThreadRunning then
  begin
    // check angles
    if not SetRotAngle(@RenderFrm.RotAlpha, edAlpha) or
       not SetRotAngle(@RenderFrm.RotBeta, edBeta) then exit;

    SetGlobalColors;

    // clear treeview
    tvData.Items.Clear;

    // dir:= edtDir.Text;
    dir:= GetDirectory('Verzeichnis whlen');
    // GetDirectory aborted
    if (dir = '') then exit; // dir:= 'D:\computer';
    edtDir.Text:= dir;
    // add slash
    if (dir[Length(dir)] <> '\') then dir:= dir + '\';

    // TReadDirThread
    FReadDirThread:= TReadDirThread.Create(true, dir, @tvData);
    TLogger.GetInstance.LogMessage('TReadDirThread created');
    FReadDirThread.FreeOnTerminate:= true;
    FReadDirThread.OnTerminate:= ReadDirThreadFinished;
    FReadDirThread.Resume;

    btnScan.Caption:= '&abort';
    FReadDirThreadRunning:= true;
  end
  else
    if Assigned(FReadDirThread) and not FReadDirThread.Terminated then
    begin
      btnScan.Enabled:= false;
      FReadDirThread.Terminate;
    end;
end;

{*
form create
do some initialising
*}
procedure TMainFrm.FormCreate(Sender: TObject);
begin
  TLogger.GetInstance.LogMessage('Application started');
  edtDir.Text:= '';
  btnScan.Caption:= '&open';
  FReadDirThread:= nil;
  FReadDirThreadRunning:= false;
  FCntFiles:= 0;
  FCntDirs:= 0;
  FProgDir:= ExtractFilePath(ParamStr(0));

  // Rendering
  RenderFrm:= TRenderFrm.Create(nil);
  if not RenderFrm.initialised then
  begin
    TLogger.GetInstance.LogMessage('OpenGL initialisation failed!');
    MessageBox(MainFrm.Handle, PChar('OpenGL initialisation failed!'), PChar('Error'), MB_OK or MB_ICONEXCLAMATION);
    Close;
  end;

  // settings
  ReadIniSettings;

  TLogger.GetInstance.LogMessage('Application initialized');
end;

{*
form close
*}
procedure TMainFrm.FormDestroy(Sender: TObject);
begin
  tvData.Items.Clear;
  WriteIniSettings;
  RenderFrm.Free;
  TLogger.GetInstance.LogMessage('Application finalization');
  TLogger.GetInstance.Free;
end;

{*
open a directory open dialog

@param  caption, the title of the dialog
@return the selected directory as string, or empty string
*}
function TMainFrm.GetDirectory(caption: string): String;
var
  bi: TBrowseInfo;
  lpBuffer: array[0..MAX_PATH + 1] of Char;
  pidlBrowse: PItemIDList;
  ppMalloc: IMalloc;
begin
  pidlBrowse:= nil;
  result:= '';

{ pidlPrograms: PItemIDList;
  pidlPrograms:= nil;
  if not SHGetSpecialFolderLocation(GetActiveWindow, CSIDL_DRIVES, pidlPrograms) = S_OK then exit;
  bi.pidlRoot:= pidlPrograms;
  ppMalloc.Free(pidlPrograms); }

  try
    ZeroMemory(@bi, sizeof(TBrowseInfo));
    with bi do
    begin
      hwndOwner:= GetActiveWindow;
      pszDisplayName:= lpBuffer;
      lpszTitle:= PChar(caption);
      ulFlags:= BIF_RETURNONLYFSDIRS;
      lpfn:= nil;
      lParam:= 0;
    end;

    pidlBrowse:= SHBrowseForFolder(bi);
    if (pidlBrowse <> nil) and (SHGetPathFromIDList(pidlBrowse, lpBuffer) = true) then
      result:= lpBuffer;
  finally
    if SHGetMalloc(ppMalloc) = NOERROR then
    begin
      ppMalloc.Free(pidlBrowse);
      ppMalloc._Release;
    end;
  end;
end;

{*
TReadDirThread execution finished
*}
procedure TMainFrm.ReadDirThreadFinished(Sender: TObject);
begin
  FReadDirThreadRunning:= false;
  btnScan.Caption:= '&open';
  btnScan.Enabled:= true;
  TLogger.GetInstance.LogMessage('TReadDirThread finished');
end;

{*
event if (sub)item of listview is drawing
*}
procedure TMainFrm.lvColTypCustomDrawSubItem(Sender: TCustomListView;
  Item: TListItem; SubItem: Integer; State: TCustomDrawState;
  var DefaultDraw: Boolean);
begin
  Sender.Canvas.Brush.Color:= HexToTColor(Item.SubItems.Strings[0]);
end;

{*
read settings from ini file
*}
procedure TMainFrm.ReadIniSettings;
var
  ini: TIniFile;
  types: TStrings;
  i: integer;
  color: String;
begin
  try
    ini:= TIniFile.Create(FProgDir + 'settings.ini');
    types:= TStringList.Create;
    ini.ReadSection('colors', types);

    // clear listview and fill it up with color data
    lvColTyp.Items.Clear;
    for i:= 0 to types.Count-1 do
    begin
      color:= ini.ReadString('colors',types[i],'ffffff');
      with lvColTyp.Items.Add do
      begin
        Caption:= types[i];
        SubItems.Add(color);
      end;
    end;
    lvColTyp.AlphaSort;         // sort table

    // rotation angles
    RenderFrm.RotAlpha:= ini.ReadFloat('common', 'rotAlpha', 45.0);
    RenderFrm.RotBeta:=  ini.ReadFloat('common', 'rotBeta', 222.5);
    edAlpha.Text:= FloatToStr(RenderFrm.RotAlpha);
    edBeta.Text:=  FloatToStr(RenderFrm.RotBeta);
  finally
    types.Free;
    ini.Free;
  end;
end;

{*
write all settings to ini file
*}
procedure TMainFrm.WriteIniSettings;
var
  ini: TIniFile;
  i: integer;
begin
  try
    ini:= TIniFile.Create(FProgDir + 'settings.ini');
    for i:= 0 to lvColTyp.Items.Count-1 do
    begin
      with lvColTyp.Items.Item[i] do
      begin
        ini.WriteString('colors', Caption, SubItems.Strings[0]);
      end;
    end;
    // rotation angles
    ini.WriteFloat('common', 'rotAlpha', RenderFrm.RotAlpha);
    ini.WriteFloat('common', 'rotBeta', RenderFrm.RotBeta);
  finally
    ini.Free;
  end;
end;

{*
keystroke handling
*}
procedure TMainFrm.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if (Key = VK_ESCAPE) then Close;
end;

{*
find colour to corresponding filetype
if undefinded, return black

@param  t, the filetype we are searching for
@return always TColor3f
*}
function TMainFrm.GetColor3fToTyp(const t: String): TColor3f;
var
  i: integer;
begin
  result:= ColorUnknown;
  for i:= 0 to lvColTyp.Items.Count-1 do
  begin
    with lvColTyp.Items.Item[i] do
    begin
      if Caption = t then
      begin
        result:= TRGBQuadToColor3f(TColorToTRGBQuad(HexToTColor(SubItems.Strings[0])));
        break;
      end;
    end;
  end;
end;

{*
add new row or delete one
*}
procedure TMainFrm.lvColTypKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if (Key = VK_DOWN) and (lvColTyp.Selected.Index = lvColTyp.Items.Count-1) then
  begin
    with lvColTyp.Items.Add do
    begin
      Caption:= '';
      SubItems.Add('FFFFFF');
    end;
  end
  else if (Key = VK_DELETE) then
  begin
    lvColTyp.Items.Delete(lvColTyp.Selected.Index);
    lvColTyp.ItemFocused.Selected:= true;
  end;
end;

procedure TMainFrm.lvColTypColumnClick(Sender: TObject;
  Column: TListColumn);
begin
  (Sender as TCustomListView).AlphaSort;
end;

{*
select new color with dialog
*}
procedure TMainFrm.lvColTypDblClick(Sender: TObject);
begin
  if colDlg.Execute then lvColTyp.Selected.SubItems[0]:= TColorToHex(colDlg.Color);
end;

{*
node click event
*}
procedure TMainFrm.tvDataClick(Sender: TObject);
var
  n: TTreeNode;
begin
  n:= tvData.Selected;
  if n = nil then exit;
  if n.Data = nil then exit;
  if RenderFrm.Showing then
  begin
    RenderFrm.PickedDir:= PGLint(n.Data)^;
    RenderFrm.Refresh;
  end;
end;


//------------------------------------------------------------------------------
// TReadDirThread
//------------------------------------------------------------------------------

constructor TReadDirThread.Create(CreateSuspended: Boolean; dir: String; const ptv: PTTreeView);
begin
  inherited Create(CreateSuspended);
  FDir:= dir;
  FPTV:= ptv;
end;

{*
destructor
*}
destructor TReadDirThread.Destroy();
begin
  FreeDI(FPDI);
  inherited Destroy;
end;

{*
thread execution starts here
*}
procedure TReadDirThread.Execute;
begin
  try
    FPTV^.Items.BeginUpdate;
    FPDI:= GetFiles(FDir, nil);
    FPTV^.Items.EndUpdate;
    Synchronize(BuildRenderData);
  except
    on e: Exception do
      MessageBox(0,PChar('Error in ' + e.ClassName + #13#10 + e.Message), 'TReadDirThread', MB_OK or MB_ICONERROR);
  end;
end;

{*
recursive freeing of directory item structure

@param  pdi, the current directory item to free
*}
procedure TReadDirThread.FreeDI(pdi: PTDirectoryItem);
var
  i,j: integer;
  pfitmp: PTFileItem;
begin
  if (pdi = nil) then exit;

  // free file items
  if pdi^.fl <> nil then
  begin
    for i:= 0 to pdi^.fl.Count-1 do
    begin
      pfitmp:= pdi^.fl.Items[i];
      Dispose(pfitmp);
    end;
    pdi^.fl.Free;
  end;

  // free dir items
  if pdi^.dl <> nil then
  begin
    for j:= 0 to pdi^.dl.Count-1 do
      FreeDI(pdi^.dl.Items[j]);

    pdi^.dl.Free;
  end;

  Dispose(pdi);
end;

{*
recursive parsing of directory by directory
with calculation of directory sizes

@param  dir, the directory we currently parse
*}
function TReadDirThread.GetFiles(const dir: String; const parent: TTreeNode): PTDirectoryItem;
var
  sr: TSearchRec;
  pfi: PTFileItem;
  pdi: PTDirectoryItem;
  offspring: TTreeNode;
begin
  result:= nil;
  if (Terminated = true) then exit;    // for thread handling
  FDir:= dir;
  Synchronize(UpdateCurrentDir);
  // treeview
  if parent = nil then
    offspring:= FPTV^.Items.AddChild(parent, dir)
  else
    offspring:= FPTV^.Items.AddChild(parent, ExtractFileName(ExcludeTrailingPathDelimiter(dir)));
  offspring.Data:= nil;

  if (FindFirst(dir + '*.*', faAnyFile, sr) = 0) then
    try
      New(result);
      result^.n:= dir;
      result^.s:= 0;
      result^.fileCnt:= 0;
      result^.dirCnt:= 0;
      result^.tn:= offspring;
      result^.dl:= TSList.Create;
      result^.fl:= TSList.Create;

      repeat
        // do not get out of the directory
        if (sr.Name = '.') or (sr.Name = '..') then continue;
        // go deeper
        if (sr.Attr and faDirectory = faDirectory) then
        begin
          pdi:= GetFiles(dir + sr.Name + '\', offspring);
          if pdi = nil then continue;           // maybe you have not enough privileges!
          if pdi^.s = 0 then                    // skip empty directory
          begin
            pdi^.tn.Free;
            FreeDI(pdi);                        // free it now
            continue;
          end;
          result^.s:= result^.s + pdi^.s;       // size of all subdirs
          result^.fileCnt:= result^.fileCnt + pdi^.fileCnt;
          result^.dirCnt:= result^.dirCnt + pdi^.dirCnt;
          result^.dl.Add(pdi);
          INC(result^.dirCnt);
        end
        else // files
        begin
          New(pfi);
          pfi^.n:= sr.Name;
          { pfi^.s:= sr.Size; // integer (2GB limit) }
          pfi^.s := sr.FindData.nFileSizeLow or (sr.FindData.nFileSizeHigh shl 32);
          result^.s:= result^.s + pfi^.s; // size of current directory
          result^.fl.Add(pfi);
          INC(result^.fileCnt);
        end;
        if (Terminated = true) then break;
      until (FindNext(sr) <> 0);
    finally
      FindClose(sr);
    end;
end;

{*
non thread safety rendering calls
*}
procedure TReadDirThread.BuildRenderData;
begin
  if not Terminated and RenderFrm.initialised then
  begin
    MainFrm.lblInfo.Caption:= 'size: ' + PrintSizeUserFriendly(FPDI^.s) +
                              ' / ' + IntToStr(FPDI^.fileCnt) + ' files, ' +
                              IntToStr(FPDI^.dirCnt) + ' dir(s)';
    RenderFrm.BuildSceneGraph(FPDI);
    If not RenderFrm.Showing then RenderFrm.Show;
  end;
end;

{*
synchronized update of label
*}
procedure TReadDirThread.UpdateCurrentDir;
begin
  MainFrm.lblInfo.Caption:= FDir;
end;

end.