{*
Node unit
some nodes for building a scene graph

@author         Philipp Hartl
@version        0.1
@license
This source file is part of BotanicalVis,
a tool to visualize huge hierarchies in 3D-space

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
*}

unit Nodes;

interface

uses
  Classes, SysUtils, ComCtrls,
  dglOpenGL, Data;

type

  TNodeType = (NODE, GROUP, TRANSFORM, SHAPE);

  /// node class
  TNode = class
  protected
    FPChildIndex: TGLuint;              /// parent childs index
    FParent: TNode;                     /// pointer to parent, for root it is nil
    FName: String;                      /// (unique) name
    FSize: Int64;                       /// size in byte
  public
    constructor Create(n: TNode = nil);
    destructor Destroy; override;
    function NodeType: TNodeType; virtual;
    procedure Render; virtual; abstract;

    property Name: String read FName write FName;
    property Parent: TNode read FParent;
    property Size: Int64 read FSize write FSize;
  end;

  {*
   group
   for e.g. some transformations and shapes
   the order of the nodes is important! (R*T != T*R)
   if this group should not affect other nodes with T, R, S, ... separator it
   *}
  TGroup = class(TNode)
  private
    FChilds: TList;                     /// list of TNodes
    FSeparator: boolean;                /// does this group affect other groups (not childs, they are always affected)
    FOffspringCnt: integer;             /// number of offspring groups (childs, not cylinder groups)
    FId: TGLuint;
    FKit: TItem;                        /// which kind of item (dir != file)
    FScale: TGLfloat;                   /// scalefactor for all childs
    FTN: TTreeNode;

    procedure Reset;

  public
    constructor Create(parent: TGroup = nil; separator: boolean = false; offspring: boolean = false);
    destructor Destroy; override;

    procedure AddChild(const n: TNode; offspring: boolean = false);
    procedure Clear;                    /// free items
    procedure FetchId;
    function GetName: String;
    function GetFirstGroup: TGroup;
    procedure InsertChild(const n: TNode; const pos: integer);
    function IsEmpty: boolean;
    function NodeType: TNodeType; override;
    procedure Render; override;         /// render each node
    procedure RenderPickedTransformsIncl(idx: TGLuint = 0);       /// inclusive (all)
    procedure RenderPickedTransformsExcl(idx: TGLuint = 0);       /// exclusive self

    property Childs: TList read FChilds;
    property ID: TGLuint Read FId Write FId;
    property Kit: TItem Read FKit Write FKit;
    property Offsprings: integer Read FOffspringCnt;
    property ScaleFactor: TGLfloat Read FScale Write FScale;
    property Separator: boolean Write FSeparator;
    property TreeNode: TTreeNode Read FTN Write FTN;
  end;

implementation

//------------------------------------------------------------------------------
// TNode
//------------------------------------------------------------------------------

constructor TNode.Create(n: TNode);
begin
  FPChildIndex:= 0;
  FParent:= n;
  FName:= '';
  FSize:= 0;
end;

destructor TNode.Destroy;
begin
  FParent:= nil;
end;

function TNode.NodeType: TNodeType;
begin
  result:= NODE;
end;

//------------------------------------------------------------------------------
// TGroup
//------------------------------------------------------------------------------

{*
constructor with maximum feelings
@param  parent, the parent group this group belongs to
@param  separator, should this group be a separator group or not (should it affect other groups?)
@param  offspring, is this group an offspring of parent?
*}
constructor TGroup.Create(parent: TGroup; separator: boolean; offspring: boolean);
begin
  inherited Create(parent);
  if parent <> nil then parent.AddChild(Self, offspring);
  FChilds:= TList.Create;
  FOffspringCnt:= 0;
  FSeparator:= separator;
  FKit:= DIR_ITEM;
  Reset;
end;

{*
free up all nodes
*}
destructor TGroup.Destroy;
begin
  Clear;
  FChilds.Free;
  inherited Destroy;
end;

{*
clear group (list) items recursively
*}
procedure TGroup.Clear;
var
  i: integer;
  n: TNode;
begin
  for i:= 0 to FChilds.Count-1 do
  begin
    n:= FChilds.Items[i];
    n.Free;
  end;
  FChilds.Clear;
  Reset;
end;

{*
append a node to this list
@param  n, a new TNode
*}
procedure TGroup.AddChild(const n: TNode; offspring: boolean = false);
begin
  n.FPChildIndex:= FChilds.Add(n);
  if offspring then INC(FOffspringCnt);
end;

function TGroup.NodeType: TNodeType;
begin
  result:= GROUP;
end;

{*
insert a node to this list on a defined position
@param  n, a new TNode
@param  pos, the position we want this node to be inserted
*}
procedure TGroup.InsertChild(const n: TNode; const pos: integer);
begin
  FChilds.Insert(pos, n);
end;

function TGroup.IsEmpty: boolean;
begin
  result:= true;
  if FChilds.Count > 0 then result:= false;
end;

{*
get the identifier of parent list index
*}
procedure TGroup.FetchId;
begin
  FId:= 0;
  if (FParent.NodeType = GROUP) then
    FId:= (FParent as TGroup).FChilds.Count - 1;
end;

function TGroup.GetName: String;
begin
  result:= FName;
  if result = '' then result:= FParent.Name;
end;

{*
return the first group of childs if found, otherwise nil
*}
function TGroup.GetFirstGroup: TGroup;
var
  i: integer;
  n: TNode;
begin
  result:= nil;
  for i:= 0 to FChilds.Count-1 do
  begin
    n:= FChilds.Items[i];
    if (n.NodeType = GROUP) then
    begin
      result:= n as TGroup;
      break;
    end;
  end;
end;

{*
all childs have to be rendered
maybe it should not affect other groups in same hierarchy
*}
procedure TGroup.Render;
var
  i: integer;
  n: TNode;
  g: TGroup;
  namepushed: boolean;          // see glPushName
begin
  if FSeparator = true then glPushMatrix;
  for i:= 0 to FChilds.Count-1 do
  begin
    namepushed:= false;
    n:= FChilds.Items[i];
    if (n.NodeType = GROUP) and (n.FName <> '') then
    begin
      g:= n as TGroup;
      if (g.Kit = DIR_ITEM) then        // normal group glPop/PushName
      begin
        glLoadName(g.FId);
      end
      else                              // subitem
      begin
        glPushName(g.FId);
        namepushed:= true;
      end;
    end;
    n.Render;
    if namepushed = true then glPopName;
  end;
  if FSeparator = true then glPopMatrix;
end;

{*
render all transform nodes of parents starting with root and finish with parent of group
*}
procedure TGroup.RenderPickedTransformsExcl(idx: TGLuint);
begin
  if (FParent <> nil) and (FParent.NodeType = GROUP) then
    (FParent as TGroup).RenderPickedTransformsIncl(FPChildIndex);
end;

{*
render all transform nodes of parents starting with root node and finish by itself
*}
procedure TGroup.RenderPickedTransformsIncl(idx: TGLuint);
var
  i: integer;
  n: TNode;
begin
  // go parents up, till root
  RenderPickedTransformsExcl(FPChildIndex);

  // now render only transform nodes (T, R, S)
  for i:= 0 to idx-1 do
  begin
    n:= FChilds.Items[i];
    if (n.NodeType = TRANSFORM) then n.Render;
  end;
end;

{*
reset important items
*}
procedure TGroup.Reset;
begin
  FId:= 0;
  FName:= '';
  FSize:= 0;
  FScale:= 1.0;
  FTN:= nil;
end;

end.