{*
Rendering unit
all rendering goes here

@author         Philipp Hartl
@version        0.1
@license
This source file is part of BotanicalVis,
a tool to visualize huge hierarchies in 3D-space

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
*}

unit Rendering;

interface

uses
  Windows, SysUtils, Math, Forms, Classes, Controls,
  dglOpenGL, Logger, Geo, Camera, Data, Shapes, Nodes, Transforms;

type

  /// render object (scenegraph, display list)
  TRObject = (SG, DL);

  {*
   rendering class
   initalise the rendering context and all openGL functions
   *}
  TRenderFrm = class(TForm)
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure FormResize(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure FormMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormMouseWheel(Sender: TObject; Shift: TShiftState;
      WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);

  public
    procedure BuildSceneGraph(const d: PTDirectoryItem);

  private
    FDC: HDC;                   /// GDI device context
    FRC: HGLRC;                 /// permanent rendering context
    FFontBase: GLuint;          /// base display list for font set
    FFrameCount: GLuint;
    FFPS: GLfloat;
    FStartTime, FTimeFreq: Int64;   /// for accurate timing
    FInit: Boolean;             /// OpenGL initialised?
    FVBOsupport: Boolean;       /// support vertex buffer objects? currently not implemented
    FCamera: TCamera;
    FSceneGraph: TGroup;        /// scenegraph tree
    FTreeDL: GLuint;            /// rendered scenegraph and saved in a display list
    /// vars for mouse handling
    FMouseMoved: Boolean;       /// mouse moved?
    FMousePressed: Boolean;     /// is currently a mouse button pressed?
    FMouseStartPos: TPoint;     /// starting position
    /// picking
    FPDIList: TList;            /// a list of pointer on pdi
    FPicked: array[0..1] of TGLuint;    /// picked identifier
    FPickedItem: TItem;         /// dir or file item picked
    /// settings
    FRenderObj: TRObject;       /// which object to render
    FShowFPS: Boolean;
    FPMode: Boolean;            /// which polygon mode (line, true = fill)
    FRotAlpha: TGLfloat;        /// the opening angle between two subbranches
    FRotBeta: TGLfloat;         /// the rotation angle along a stem, phyllotaxis

    procedure StartOverlay;     /// switch to 2D ortho view
    procedure EndOverlay;       /// reset
    procedure BuildFont(const fontname: String = 'Courier New');
    procedure ErrorHandling;    /// handle opengl errors
    procedure Display;             /// all drawing goes here
    procedure InitGL;           /// initialise rendering context
    function IsExtensionSupported(const name: String): Boolean; /// OpenGL ext available?
    procedure PrintText(const x, y: GLushort; const text: String);    /// print text to rendering context
    procedure Reshape;          /// viewport and projection update
    procedure Picking;          /// render in selection mode and test for hits
    procedure Timing;           /// FPS counter

    // botanic stuff
    procedure phiball(const parent: TGroup; const d: PTDirectoryItem; const top, bot: TGLfloat);
    procedure dtree1(const parent: TGroup; const d: PTDirectoryItem);
    procedure dtree2(const parent: TGroup; const d: PTDirectoryItem; const e: double);

  public
    property initialised: Boolean read FInit;
    property RotAlpha: TGLfloat read FRotAlpha Write FRotAlpha;
    property RotBeta: TGLfloat read FRotBeta Write FRotBeta;
    property PickedDir: TGLuint Write FPicked[0];       // set picked dir

  end;

var
  RenderFrm: TRenderFrm;

implementation

uses Main;

{$R *.dfm}

//------------------------------------------------------------------------------
// TRenderFrm
//------------------------------------------------------------------------------

{*
initialise opengl
*}
procedure TRenderFrm.FormCreate(Sender: TObject);
begin
  TLogger.GetInstance.LogMessage('TRenderFrm created');
  FShowFPS:= false;
  FFrameCount:= 0;
  FFPS:= 0.0;
  FMousePressed:= false;
  FSceneGraph:= TGroup.Create;
  FPDIList:= TList.Create;
  FPickedItem:= DIR_ITEM;
  FRenderObj:= DL;
  FPMode:= true;
  FRotAlpha:= 45;
  FRotBeta:= 222.5;             // 360 / RotGamma; RotGamma = (1 + Sqrt(5)) / 2; = 1.618034;
  FInit:= InitOpenGL;           // initialise openGL functions, @see dglOpenGL
  If FInit then InitGL;
end;

{*
free up
*}
procedure TRenderFrm.FormDestroy(Sender: TObject);
begin
  FSceneGraph.Free;
  FPDIList.Free;
  FCamera.Free;
  glDeleteLists(FFontBase, 96);
  glDeleteLists(FTreeDL, 1);
  FreeBaseShapes;
  FreeGLUQuadricObj;
  DeactivateRenderingContext;
  wglDeleteContext(FRC);
  ReleaseDC(Self.Handle, FDC);
  TLogger.GetInstance.LogMessage('TRenderFrm freed');
end;

{*
create font display list

@param  fontname, the font we want to use (default: Courier New)
*}
procedure TRenderFrm.BuildFont(const fontname: String);
var
 Font: HFONT;
begin
  FFontBase:= glGenLists(96);
  Font:= CreateFont(16, 0, 0, 0, FW_MEDIUM, 0, 0, 0, ANSI_CHARSET, OUT_TT_PRECIS,
         CLIP_DEFAULT_PRECIS, ANTIALIASED_QUALITY, FF_DONTCARE or DEFAULT_PITCH,
         PChar(fontname));
  SelectObject(FDC, Font);
  wglUseFontBitmaps(FDC, 0, 256, FFontBase);
  DeleteObject(Font);
end;

{*
build up botanic tree
currently use immediate rendering mode => display list
automatically deletes the old list if called again
*}
procedure TRenderFrm.BuildSceneGraph(const d: PTDirectoryItem);
begin
  if d = nil then exit;
  // clear previous data
  FSceneGraph.Clear;
  FPDIList.Clear;
  // reset picking data to default
  FPicked[0]:= 0;
  FPicked[1]:= 0;
  FPickedItem:= DIR_ITEM;

  dTree1(FSceneGraph, d);
  if FSceneGraph.IsEmpty then exit;

  // make dl too, automatically deletes the old list if called again
  glNewList(FTreeDL, GL_COMPILE);
    FSceneGraph.Render;
  glEndList;
end;

{*
phiball

@param  d, the directory item with files
@param  top, how much is the sphere in its upper hemisphere available
@param  bot, lower hemisphere, do not obscure file cones in the stem
@see    Botanical Visualization of Huge Hierarchies by Kleiberg, Wetering, Wijk
*}
procedure TRenderFrm.phiball(const parent: TGroup; const d: PTDirectoryItem; const top, bot: TGLfloat);
var
  f: PTFileItem;
  h, dts, c, t: TGLfloat;
  i: integer;
  size: Int64;
  g: TGroup;
  sf: TGLfloat;
  s: TShape;
  ext: String;
begin
  // recalc size of directory with files, maybe it does not compare to d^.s
  size:= 0;
  for i:= 0 to d^.fl.Count-1 do
  begin
    f:= d^.fl.Items[i];
    size:= size + f^.s;
  end;
  if d^.s <> size then d^.s:= size;
  if d^.s = 0 then exit;

  // properties
  parent.ID:= FPDIList.Add(parent);
  parent.Name:= d^.n;
  parent.Size:= d^.s;
  parent.TreeNode:= d^.tn;
  sf:= parent.ScaleFactor;
  if d^.tn.Data = nil then d^.tn.Data:= @parent.ID;     // reference

  parent.AddChild(TQSCylinder.Create(sf * 2.0, sf * 2.0, sf * 8.0, 16, 1));  // quadric cylinder
  // move on top of stem (stemheight + Sqrt(Sqr(sphereradius) - Sqr(stemradius)))
  parent.addChild(TTTranslate.Create(Vector3f(0, sf * 11.4641, 0)));
  parent.AddChild(TQSSphere.Create(sf * 4.0, 16, 16));

  h:= top;
  for i:= 0 to d^.fl.Count-1 do
  begin
    f:= d^.fl.Items[i];
    dts:= f^.s / d^.s;          // dt size
    c:= Sqrt(dts);              // hypotenuse of cone
    t:= Sqrt(dts + 1);          // Sqrt(c + 1)

    parent.addChild(TTRotate.Create(RotBeta, Vector3f_UnitY));          // R(y)
    // create for each cone a new separator group and attach it to parent
    g:= TGroup.Create(parent, true);
    g.Kit:= FILE_ITEM;          // file properties
    g.FetchId;
    g.Name:= f^.n;
    g.Size:= f^.s;
    g.addChild(TTRotate.Create(-RadToDeg(ArcSin(h)), Vector3f_UnitX));  // R(x)
    g.addChild(TTTranslate.Create(Vector3f(0, 0, sf * 4.0 / t)));       // move B out
    s:= TQSCone.Create(sf * 4.0 * c / t, 0, sf * 4.0 * dts / t, 16, 1); // cone
    ext:= ExtractFileExt(f^.n);
    s.Color:= MainFrm.GetColor3fToTyp(Copy(ext, 2, Length(ext)-1));     // get color match
    g.AddChild(s);

    h:= h - ((top - bot) * dts);
  end;
end;

{*
dtree1
build up a display list (tree) by traversing the directory structure
special case of dtree2 (epsilon=0)

@see    Botanical Visualization of Huge Hierarchies by Kleiberg, Wetering, Wijk
*}
procedure TRenderFrm.dtree1(const parent: TGroup; const d: PTDirectoryItem);
var
  dts1, dts2, sqdts1, sqdts2, maxsqdts: single;
  d1: PTDirectoryItem;
  g: TGroup;
  h, sf: TGLfloat;
begin
  if d = nil then exit;

  if d^.dl.PCount = 0 then              // no more directories, make phiball (leaves)
  begin
    if d^.s > 0 then phiball(parent, d, 1, -0.7071);    // the space round stem should not be taken
    exit;
  end;

  // set each group only once!
  if (parent.Name = '') then
  begin
    parent.ID:= FPDIList.Add(parent);   // identifier for picking
    parent.Name:= d^.n;                 // directory name
    parent.Size:= d^.s;                 // directory size (maybe left size)
    parent.TreeNode:= d^.tn;            // treenode pointer for reference
  end;

  if d^.tn.Data = nil then d^.tn.Data:= @parent.ID;     // a ref on current group

//  d1:= d^.dl.PopFirst;                // always pop of the first element of the list
  d1:= d^.dl.Next;
  dts1:= d1^.s / d^.s;                  // calculate the delta ratios of the childs
  dts2:= 1 - dts1;
  sqdts1:= Sqrt(dts1);
  sqdts2:= Sqrt(dts2);
  d^.s:= d^.s - d1^.s;                  // adapt new size of continuing branch
  h:= Sin(DegToRad(RotAlpha * dts2)) * sqdts1 * 4.0;            // calc the height of the first rotated stem
  sf:= parent.ScaleFactor;
  maxsqdts:= Max(sqdts1, sqdts2);

  // for both sub-trees
  parent.AddChild(TQSCylinder.Create(sf * 2.0, sf * 2.0 * maxsqdts, sf * 8.0, 16, 1));    // base stem
  // move to the top of the current stem, avoiding holes by lowering of height of first stem
  parent.addChild(TTTranslate.Create(Vector3f(0, sf * (8.0 - h), 0)));
  parent.addChild(TTRotate.Create(RotBeta, Vector3f_UnitY));    // always rotate a little bit along the height axis of the stem (phyllotaxis)

  // separated subgroup "left-branch"
  g:= TGroup.Create(parent, true);
  g.ScaleFactor:= sqdts1 * sf;
  if sqdts1 < 1 then                    // not only one sub-branch?
  begin
    g.addChild(TTTranslate.Create(Vector3f(0, 0, sf * 2.0 * maxsqdts)));     // move radius out to the edge of the stem
    g.addChild(TTRotate.Create(RotAlpha * dts2, Vector3f_UnitX));            // opening angle between two sub-branches
    g.addChild(TTTranslate.Create(Vector3f(0, 0, sf * -2.0 * sqdts1)));      // move the rotated sub-branch (stem) r/2 back
  end;
  dtree1(g, d1);                        // follow left-branch

  // separated subgroup "right-branch"
  // never create here a separated group, till it will overflow the modelview stack
  g:= TGroup.Create(parent);
  g.ScaleFactor:= sqdts2 * sf;
  if sqdts2 > 0 then
  begin
    g.addChild(TTTranslate.Create(Vector3f(0, 0, sf * -2.0 * maxsqdts)));
    g.addChild(TTRotate.Create(-RotAlpha * dts1, Vector3f_UnitX));
    g.addChild(TTTranslate.Create(Vector3f(0, 0, sf * 2.0 * sqdts2)));
  end;
  dtree1(g, d);
end;

{*
dtree2
similar to dtree1, but we are not limited on a binary tree

@see    Botanical Visualization of Huge Hierarchies by Kleiberg, Wetering, Wijk
*}
procedure TRenderFrm.dtree2(const parent: TGroup; const d: PTDirectoryItem; const e: double);
const epsilon: single = 1.0;            // [0,1], 0 only one directory per stem, 1 all directories are splitted off
var
  g: TGroup;
  f: double;
  dts1, dts2, sqdts1, sqdts2: single;
  d1: PTDirectoryItem;
begin
  if d = nil then exit;

  if d^.dl.PCount = 0 then              // no more directories, make phiball (leaves)
  begin
    if d^.s > 0 then phiball(parent, d, 1, -0.7071);    // the space round stem should not be taken
    exit;
  end;

  // set each group only once!
  if (parent.Name = '') then
  begin
    parent.ID:= FPDIList.Add(parent);   // identifier for picking
    parent.Name:= d^.n;                 // directory name
    parent.Size:= d^.s;                 // directory size (maybe left size)
  end;

  if e <= 0 then
  begin
    f:= epsilon * d^.s;
    g:= TGroup.Create(parent, true);    // subgroup for cylinder, base stem
    g.addChild(TTRotate.Create(-90, Vector3f_UnitX));
    g.addChild(TQSCylinder.Create(2.0, 2.0, 8.0, 16, 1));
    parent.addChild(TTTranslate.Create(Vector3f(0, 8.0, 0)));           // move to the top of the current stem
  end
  else f:= e;

  d1:= d^.dl.Next;
  dts1:= d1^.s / d^.s;                  // calculate the delta ratios of the childs
  dts2:= 1 - dts1;
  sqdts1:= Sqrt(dts1);
  sqdts2:= Sqrt(dts2);

  d^.s:= d^.s - d1^.s;
  f:= f - d1^.s;

  // for both sub-trees
  parent.addChild(TTRotate.Create(RotBeta, Vector3f_UnitY));            // always rotate a little bit along the height axis of the stem (phyllotaxis)

  // separated subgroup "left-branch"
  g:= TGroup.Create(parent, true, true);
  if sqdts1 < 1 then                    // not only one sub-branch?
  begin
    g.addChild(TTTranslate.Create(Vector3f(0, 0, 2.0)));                // move radius out to the edge of the stem
    g.addChild(TTRotate.Create(RotAlpha * dts2, Vector3f_UnitX));       // opening angle between two sub-branches
    g.addChild(TTTranslate.Create(Vector3f(0, 0, -2.0 * sqdts1)));      // move the rotated sub-branch (stem) r/2 back
    g.addChild(TTScale.Create(Vector3f(sqdts1, sqdts1, sqdts1)));       // scale it in respect of sqdts1
  end;
  g.addChild(TQSCoordAxes.Create);
  dtree2(g, d1, 0);

  // separated subgroup "right-branch"
  // never create here a separated group, till it will overflow the modelview stack
  g:= TGroup.Create(parent, false, true);
  if sqdts2 > 0 then
  begin
    g.addChild(TTTranslate.Create(Vector3f(0, 0, -2.0)));
    g.addChild(TTRotate.Create(-RotAlpha * dts1, Vector3f_UnitX));
    g.addChild(TTTranslate.Create(Vector3f(0, 0, 2.0 * sqdts2)));
    g.addChild(TTScale.Create(Vector3f(sqdts2, sqdts2, sqdts2)));
  end;
  g.addChild(TQSCoordAxes.Create);
  dtree2(g, d, f);
end;

{*
proceed all drawing
*}
procedure TRenderFrm.Display;
var
  g: TGroup;
begin
  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
  glColor3f(1.0, 1.0, 1.0);
  glLoadIdentity;

  FCamera.LookAt;               // camera lookat matrix

  // 3D tree (static or dynamic)
  glPushMatrix;
    glEnable(GL_LIGHTING);
    if FRenderObj = SG then FSceneGraph.Render
    else glCallList(FTreeDL);
    glDisable(GL_LIGHTING);
  glPopMatrix;

  // handle picked object
  g:= FSceneGraph;
  if FPicked[0] < FPDIList.Count then
  begin
    g:= FPDIList.Items[FPicked[0]];       // get group pointer (dir_item)

    // treeview
    if g.TreeNode <> nil then
    begin
      g.TreeNode.Expanded:= true;
      g.TreeNode.MakeVisible;
      g.TreeNode.Selected:= true;
    end;

    glPushMatrix;
      // render all previous transformations to get the right position of picked object
      (g.GetFirstGroup).RenderPickedTransformsExcl;
      glCallList(BSCoordAxesDL);  // here you can draw what you want, to emphasize the object
    glPopMatrix;
  end;

  // text
  StartOverlay;                 // switch to 2D Overlay
  PrintText(5, 15, 'Directory: ' + g.Name);
  PrintText(5, 30, 'Size: ' + PrintSizeUserFriendly(g.Size));
  // some more informations are possible (fileCnt, dirCnt, ...)
  if FShowFPS then PrintText(5, ClientHeight - 7, FloatToStr(FFPS) + ' fps');
  if (FPickedItem = FILE_ITEM) then
  begin
    if (FPicked[1] < g.Childs.Count) then
    begin
      g:= g.Childs.Items[FPicked[1]];     // get group pointer (file_item)
      PrintText(5, 60, 'File: ' + g.Name);
      PrintText(5, 75, 'Size: ' + PrintSizeUserFriendly(g.Size));
    end;
  end;
  EndOverlay;

  SwapBuffers(FDC);

  ErrorHandling;                // you make me crazy
  Timing;                       // recalc timer
end;

{*
redraw a small window area with picking amount
note that identifier starts at 1
0 should not be used as identifier
*}
procedure TRenderFrm.Picking;
var
  buffer: array[0..255] of GLuint;
  vvp: TVector4i;
  hits, i: TGLint;
  pbuf: PGLuint;
  minZ, names: TGLuint;
begin
  glSelectBuffer(256, PGLuint(@buffer));
  glRenderMode(GL_SELECT);              // change to selection mode
  glInitNames;
  glPushName(0);

  // setup picking matrix and render to selection buffer
  glMatrixMode(GL_PROJECTION);          // save our projection
  glPushMatrix;
  glLoadIdentity;

  glGetIntegerv(GL_VIEWPORT, PGLint(@vvp));     // get current viewport
  gluPickMatrix(FMouseStartPos.X, vvp[3] - FMouseStartPos.Y, 1.0, 1.0, vvp);
  gluPerspective(FCamera.Fovy, FCamera.AspectRatio, FCamera.NearPlane, FCamera.FarPlane);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity;

  // render (only things that could be picked)
  FCamera.LookAt;                       // camera lookat matrix
  glCallList(FTreeDL);

  glMatrixMode(GL_PROJECTION);          // restore our projection
  glPopMatrix;
  glMatrixMode(GL_MODELVIEW);

  hits:= glRenderMode(GL_RENDER);

  {*
  the buffer looks like this, the fields:
  0 ... the number of names it contains
  1 ... minZ
  2 ... maxZ
  3 and x more numbers ... name (optional!, empty if first field =0)
  *}

  FPickedItem:= DIR_ITEM;
  minZ:= High(TGLuint);         // maximum possible value
  pbuf:= PGLuint(@buffer);      // first field
  for i:= 0 to hits-1 do        // get closest picked object by traversing the z-buffer
  begin
    names:= pbuf^;              // 0 ... number of names it contains
    Inc(pbuf);                  // 1 ... minZ
    if (pbuf^ < minZ) and (names > 0) then
    begin
      minZ:= pbuf^;
      FPicked[0]:= PGLuint(TGLuint(pbuf) + 8)^;         // 3 ... first name (SizeOf(TGLuint) * 2 = 8)
      if names > 1 then                                 // be aware of overlapping shapes!
      begin
        FPicked[1]:= PGLuint(TGLuint(pbuf) + 12)^;      // 4 ... second name (SizeOf(TGLuint) * 3 = 12)
        FPickedItem:= FILE_ITEM;
      end;
    end;
    Inc(pbuf, names + 2);
  end;
  Display;
end;

{*
handle openGL errors
*}
procedure TRenderFrm.ErrorHandling;
var
  errCode: TGLEnum;
begin
  errCode:= glGetError;
  if errCode <> GL_NO_ERROR then
    TLogger.GetInstance.LogMessage('Error: ' + gluErrorString(errCode));
end;

{*
initialise rendering context

@param  handle, the rendering handle
*}
procedure TRenderFrm.InitGL;
const
  LightPos: array[0..3] of TGLfloat = (0, 1.0, 1.0, 1.0);
//  var stackdepth: GLuint;
begin
  // take handle of formular
  FDC:= GetDC(Self.Handle);
  // create rendering context (32bit colour depth, 24bit depth buffer, double buffering)
  FRC := CreateRenderingContext(FDC, [opDoubleBuffered], 32, 24, 0, 0, 0, 0);
  ActivateRenderingContext(FDC, FRC);

  // some short information
  TLogger.GetInstance.LogMessage('OpenGL ver. ' + glGetString(GL_VERSION));
  TLogger.GetInstance.LogMessage('* Vendor: ' + glGetString(GL_VENDOR));
  TLogger.GetInstance.LogMessage('* Renderer: ' + glGetString(GL_RENDERER));
//  TLogger.GetInstance.LogMessage('* Extensions: ' + PChar(glGetString(GL_EXTENSIONS)));
//  glGetIntegerv(GL_MAX_NAME_STACK_DEPTH, @stackdepth);
//  TLogger.GetInstance.LogMessage('* Name Stack Depth: ' + IntToStr(stackdepth));

  FVBOsupport:= IsExtensionSupported('GL_ARB_vertex_buffer_object');

  // init camera
  FCamera:= TCamera.Create(Vector3f(20, 10, 20), Vector3f(0,10,0), Vector3f_UnitY);
  // FCamera:= TCamera.Create;

  // black background colour
  glClearColor(0, 0, 0, 0);
  glClearDepth(1.0);

  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LEQUAL);
  glShadeModel(GL_SMOOTH);

  // backface culling
  glEnable(GL_CULL_FACE);
  glCullFace(GL_BACK);
  // the most correct, or highest quality, option should be chosen
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

  // light, do not forget to switch on the light
  glLightfv(GL_LIGHT0, GL_POSITION, PGLfloat(@LightPos));
  glEnable(GL_LIGHT0);
  glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE);
  glEnable(GL_COLOR_MATERIAL);

  // build static display lists
  BuildFont;
  InitGLUQuadricObj;
  BuildBaseShapes;
  FTreeDL:= glGenLists(1);

  // initialise timer settings
  QueryPerformanceFrequency(FTimeFreq);
  QueryPerformanceCounter(FStartTime);
end;

{*
is extension supported by this openGL version?

@param  name, the name of extension we are looking for
@return true if extension is supported, otherwise false
*}
function TRenderFrm.IsExtensionSupported(const name: String): Boolean;
var
  ext: String;
begin
  // get all available extensions
  ext:= glGetString(GL_EXTENSIONS);
  result:= pos(name, ext) > 0;
end;

{*
2D ortho view with origin in upper left corner
depth range (-1,1)
*}
procedure TRenderFrm.StartOverlay;
begin
  glMatrixMode(GL_PROJECTION);
  glPushMatrix;
  glLoadIdentity;
  gluOrtho2D(0, Self.ClientWidth, Self.ClientHeight, 0);
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix;
  glLoadIdentity;
  glListBase(FFontBase);
end;

{*
print text to rendering context
*}
procedure TRenderFrm.PrintText(const x, y: GLushort; const text: String);
begin
  if (text = '') then exit;

  glRasterPos2i(x, y);
  glCallLists(Length(text), GL_UNSIGNED_BYTE, PChar(text));
end;

{*
end 2D-overlay
*}
procedure TRenderFrm.EndOverlay;
begin
  glMatrixMode(GL_PROJECTION);
  glPopMatrix;
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix;
end;

{*
reshape viewport and projection in respect of pixel width x height
standard 3D viewing
*}
procedure TRenderFrm.Reshape;
begin
  FCamera.AspectRatio:= Self.ClientWidth / Self.ClientHeight;
  glViewport(0, 0, Self.ClientWidth, Self.ClientHeight);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity;
  if FCamera.ProjectionType = PT_PERSPECTIVE then
    gluPerspective(FCamera.Fovy, FCamera.AspectRatio, FCamera.NearPlane, FCamera.FarPlane)
  else
    glOrtho(-FCamera.AspectRatio, FCamera.AspectRatio, -FCamera.AspectRatio, FCamera.AspectRatio, FCamera.NearPlane, FCamera.FarPlane);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity;
end;

{*
recalculate timer parameters and FPS
do not calculate each frame!
*}
procedure TRenderFrm.Timing;
var
  tmp: double;
  currenttime: Int64;
begin
  INC(FFrameCount);
  QueryPerformanceCounter(currenttime);
  tmp:= currenttime - FStartTime;
  if (tmp > 0.5) then
  begin
    FFPS:= FFrameCount * FTimeFreq / tmp;
    FStartTime:= currenttime;
    FFrameCount:= 0;
  end;
  FFPS:= round(FFPS);
end;

{*
resize window
*}
procedure TRenderFrm.FormResize(Sender: TObject);
begin
  Reshape;
  Display;
end;

{*
redraw window
*}
procedure TRenderFrm.FormPaint(Sender: TObject);
begin
  Display;
end;

{*
keystroke handling
*}
procedure TRenderFrm.FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  case key of
    VK_ESCAPE: Close;                           // 27
    ord('F'):                                   // fps
    begin
      FShowFPS:= not FShowFPS;
      Display;
    end;
    ord('P'):                                   // polygon mode
    begin
      FPMode:= not FPMode;
      if FPMode then  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)
      else glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
      Display;
    end;
    ord('R'):                                   // SG, DL
    begin
      if FRenderObj = SG then FRenderObj:= DL
      else FRenderObj:= SG;
      Display;
    end;
  end;
end;

{*
mouse button pressed handling
*}
procedure TRenderFrm.FormMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  FMouseMoved:= false;
  FMousePressed:= true;
  FMouseStartPos:= Point(X, Y);
end;

{*
mouse moving
mbLeft ... paning
mbRight ... rotating about x and y
*}
procedure TRenderFrm.FormMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var
  dx, dy: GLshort;
begin
  if not FMousePressed then exit;
  FMouseMoved:= true;

  // delta distance
  dx:= X - FMouseStartPos.X;
  dy:= Y - FMouseStartPos.Y;

  // left mouse button
  if (ssLeft in Shift) then
  begin
    if dx <> 0 then FCamera.RotateGlobal(dx, 0, 1, 0);                  // rotation y World
  end
  // middle
  else if (ssMiddle in Shift) then
  begin
    if dy <> 0 then FCamera.TranslateLocal(dy * 0.5, 0, 0, -1);         // moving in z
  end
  // right button
  else if (ssRight in Shift) then
  begin
    if dx <> 0 then FCamera.TranslateLocal(dx * 0.2, -1, 0, 0);         // x-pan
    if dy <> 0 then FCamera.translateLocal(dy * 0.2, 0, 1, 0);          // y-pan
  end;

// if dy <> 0 then FCamera.rotateLocal(1, dy, 0, 0);              // rotation x Camera

  FMouseStartPos:= Point(X, Y);         // new start position
  Display;                              // redraw scene
end;

{*
reset mouse pressed variable
*}
procedure TRenderFrm.FormMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  FMousePressed:= false;
  if not FMouseMoved then Picking;
end;

{*
translating in z
*}
procedure TRenderFrm.FormMouseWheel(Sender: TObject; Shift: TShiftState;
  WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
begin
  FCamera.TranslateLocal(WheelDelta / 240, 0, 0, 1);
  Display;
end;

end.