{*
Shapes unit
cone, cylinder, sphere

@author         Philipp Hartl
@version        0.1
@license
This source file is part of BotanicalVis,
a tool to visualize huge hierarchies in 3D-space

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
*}

unit Shapes;

interface

uses
  dglOpenGL, Nodes, Geo;

type

  /// shape types
  TShapeType = (CONE, CYLINDER, SPHERE, COORDAXES);

  /// shape class
  TShape = class(TNode)
  protected
    FColor: TColor3f;
  public
    function NodeType: TNodeType; override;
    property Color: TColor3f Read FColor Write FColor;
  end;

  /// qluquadric coord axes
  TQSCoordAxes = class(TShape)
  public
    procedure render; override;
  end;

  /// shape cylinder
  TSCylinder = class(TShape)
  protected
    FBase, FTop, FHeight: TGLfloat;
    FSlices, FStacks: TGLint;
  public
    constructor Create(base, top, height: TGLfloat; slices, stacks: TGLint);
    property RadiusBottom: TGLfloat read FBase;
    property RadiusTop: TGLfloat read FTop;
    property Height: TGLfloat read FHeight;
  end;

  /// gluquadric cylinder
  TQSCylinder = class(TSCylinder)
  public
    procedure render; override;
  end;

  /// shape cone
  TSCone = class(TSCylinder)
  end;

  /// gluquadric cone
  TQSCone = class(TSCone)
  public
    procedure render; override;
  end;

  /// shape sphere
  TSSphere = class(TShape)
  protected
    FRadius: TGLfloat;
    FSlices, FStacks: TGLint;
  public
    constructor Create(radius: TGLfloat; slices, stacks: TGLint);
    property Radius: TGLfloat read FRadius;
  end;

  /// gluquadric sphere
  TQSSphere = class(TSSphere)
  public
    procedure render; override;
  end;

  /// some basic stuff
  procedure BuildBaseShapes;
  procedure FreeBaseShapes;
  procedure InitGLUQuadricObj;
  procedure FreeGLUQuadricObj;

var
  pqobj: PGLUQuadricObj;                /// global quadric object
  BSCoordAxesDL: TGLuint;               /// coord axes display list

implementation

//------------------------------------------------------------------------------

{*
build static base shapes using display lists
Cone, CoordinateAxes, Cylinder, Sphere
*}
procedure BuildBaseShapes;
begin
  // coordinate axes
  BSCoordAxesDL:= glGenLists(1);
  glNewList(BSCoordAxesDL, GL_COMPILE);
    // x = red
    glColor3f(1.0, 0, 0);
    glBegin(GL_LINES);
      glVertex3s(0, 0, 0);
      glVertex3s(10, 0, 0);
    glEnd;
    // y = green
    glColor3f(0, 1.0, 0);
    glBegin(GL_LINES);
      glVertex3s(0, 0, 0);
      glVertex3s(0, 10, 0);
    glEnd;
    // z = blue
    glColor3f(0, 0, 1.0);
    glBegin(GL_LINES);
      glVertex3s(0, 0, 0);
      glVertex3s(0, 0, 10);
    glEnd;
    // reset color and light
    glColor3f(1.0, 1.0, 1.0);
  glEndList;
end;

{*
free base shapes memory
*}
procedure FreeBaseShapes;
begin
  glDeleteLists(BSCoordAxesDL, 1);
end;

{*
filled draw style
smooth shaded
*}
procedure InitGLUQuadricObj;
begin
  pqobj:= gluNewQuadric;
end;

procedure FreeGLUQuadricObj;
begin
  gluDeleteQuadric(pqobj);
end;

//------------------------------------------------------------------------------
// TQSCoordAxes
//------------------------------------------------------------------------------

procedure TQSCoordAxes.render;
begin
  glCallList(BSCoordAxesDL);
end;

//------------------------------------------------------------------------------
// TQSCone
//------------------------------------------------------------------------------

procedure TQSCone.render;
begin
  glColor3fv(PGLfloat(@FColor));
  gluCylinder(pqobj, FBase, 0, FHeight, FSlices, FStacks);
end;

//------------------------------------------------------------------------------
// TQSCylinder
//------------------------------------------------------------------------------

procedure TQSCylinder.render;
begin
  glColor3fv(PGLfloat(@FColor));     // using GL_COLOR_MATERIAL, GL_DIFFUSE, GL_FRONT_AND_BACK
  glPushMatrix;
    glRotatef(-90.0, 1.0, 0, 0);
    gluCylinder(pqobj, FBase, FTop, FHeight, FSlices, FStacks);
    // bottom disk
    glPushMatrix;
      glRotatef(180, 1.0, 0, 0);
      gluDisk(pqobj, 0, FBase, FSlices, FStacks);
    glPopMatrix;
    // top disk
    glPushMatrix;
      glTranslatef(0, 0, FHeight);
      gluDisk(pqobj, 0, FTop, FSlices, FStacks);
    glPopMatrix;
  glPopMatrix;
end;

//------------------------------------------------------------------------------
// TQSphere
//------------------------------------------------------------------------------

procedure TQSSphere.render;
begin
  glColor3fv(PGLfloat(@FColor));        // using GL_COLOR_MATERIAL, GL_DIFFUSE, GL_FRONT_AND_BACK
  gluSphere(pqobj, FRadius, FSlices, FStacks);
end;

//------------------------------------------------------------------------------
// TShape
//------------------------------------------------------------------------------

function TShape.NodeType: TNodeType;
begin
  result:= SHAPE;
end;

//------------------------------------------------------------------------------
// TSCylinder
//------------------------------------------------------------------------------

constructor TSCylinder.Create(base, top, height: TGLfloat; slices, stacks: TGLint);
begin
  FBase:=   base;
  FTop:=    top;
  FHeight:= height;
  FSlices:= slices;
  FStacks:= stacks;
  FColor:=  ColorCylinder;
end;

//------------------------------------------------------------------------------
// TSSphere
//------------------------------------------------------------------------------

constructor TSSphere.Create(radius: TGLfloat; slices, stacks: TGLint);
begin
  FRadius:= radius;
  FSlices:= slices;
  FStacks:= stacks;
  FColor:=  ColorSphere;
end;

end.