{*
Transforms unit
transformations (translate, rotate and scale)

@author         Philipp Hartl
@version        0.1
@license
This source file is part of BotanicalVis,
a tool to visualize huge hierarchies in 3D-space

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.

*}

unit Transforms;

interface

uses
  dglOpenGL, Geo, Nodes;

type

  /// transformation types
  TTransformType = (ROTATE, SCALE, TRANSLATE);

  /// transform class
  TTransform = class(TNode)
  public
    constructor Create; virtual; abstract;
    function NodeType: TNodeType; override;
    function getType: TTransformType; virtual; abstract;
  end;

  /// rotation transform class
  TTRotate = class(TTransform)
  private
    FAngle: TGLfloat;           /// rotation angle []
    FAxis: TVector3f;           /// rotation axis

  public
    constructor Create; overload; override;
    constructor Create(const alpha: TGlfloat; const v: TVector3f); reintroduce; overload;
    function getType: TTransformType; override;
    procedure render; override;

    property alpha: TGLfloat read FAngle write FAngle;
    property x: TGLfloat read FAxis[0] write FAxis[0];
    property y: TGLfloat read FAxis[1] write FAxis[1];
    property z: TGLfloat read FAxis[2] write FAxis[2];
    property v: TVector3f read FAxis write FAxis;
  end;

  /// scaling transform class
  TTScale = class(TTransform)
  private
    FScale: TVector3f;
  public
    constructor Create; overload; override;
    constructor Create(const v: TVector3f); reintroduce; overload;
    function getType: TTransformType; override;
    procedure render; override;

    property x: TGLfloat read FScale[0] write FScale[0];
    property y: TGLfloat read FScale[1] write FScale[1];
    property z: TGLfloat read FScale[2] write FScale[2];
    property v: TVector3f read FScale write FScale;
  end;

  /// translation transform class
  TTTranslate = class(TTransform)
  private
    FTranslate: TVector3f;
  public
    constructor Create; overload; override;
    constructor Create(const v: TVector3f); reintroduce; overload;
    function getType: TTransformType; override;
    procedure render; override;

    property x: TGLfloat read FTranslate[0] write FTranslate[0];
    property y: TGLfloat read FTranslate[1] write FTranslate[1];
    property z: TGLfloat read FTranslate[2] write FTranslate[2];
    property v: TVector3f read FTranslate write FTranslate;
  end;

implementation

//------------------------------------------------------------------------------
// TTransform
//------------------------------------------------------------------------------

function TTransform.NodeType: TNodeType;
begin
  result:= TRANSFORM;
end;

//------------------------------------------------------------------------------
// TTRotate
//------------------------------------------------------------------------------

constructor TTRotate.Create;
begin
  FAngle:= 0.0;
  FAxis:= Vector3f_Zero;
end;

constructor TTRotate.Create(const alpha: TGLfloat; const v: TVector3f);
begin
  FAngle:= alpha;
  FAxis:= v;
end;

function TTRotate.getType: TTransformType;
begin
  result:= ROTATE;
end;

procedure TTRotate.render;
begin
  glRotatef(alpha, x, y, z);
end;

//------------------------------------------------------------------------------
// TTScale
//------------------------------------------------------------------------------

constructor TTScale.Create;
begin
  FScale:= Vector3f_UScale;
end;

constructor TTScale.Create(const v: TVector3f);
begin
  FScale:= v;
end;

function TTScale.getType: TTransformType;
begin
  result:= SCALE;
end;

procedure TTScale.render;
begin
  glScalef(x, y, z);
end;

//------------------------------------------------------------------------------
// TTTranslate
//------------------------------------------------------------------------------

constructor TTTranslate.Create;
begin
  FTranslate:= Vector3f_Zero;
end;

constructor TTTranslate.Create(const v: TVector3f);
begin
  FTranslate:= v;
end;

function TTTranslate.getType: TTransformType;
begin
  result:= TRANSLATE;
end;

procedure TTTranslate.render;
begin
  glTranslatef(x, y, z);
end;

end.